/********************************************************************\

  Name:         scs_210_nhq20xm.c
  Created by:   Andreas Suter (based on Stefan Ritt's scs_210.c)

  Contents:     Application specific (user) part of
                Midas Slow Control Bus protocol 
                for SCS-210 RS232 node
				running software specific for the communicatoin
				with the iseg NHQ20xm high voltage modules

  $Log: scs_210_nhq20xm.c,v $
  Revision 1.4  2005/10/04 11:46:00  nemu
  changed cmd handling

  Revision 1.1  2005/05/11 09:11:40  nemu
  initial revision


\********************************************************************/

#include <stdio.h>
#include <string.h> 
#include <intrins.h>
#include "../mscbemb.h"

//--- define commands -----------------------------------------------
#define NHQ_SYNC 0  //!< synchnoization command tag
#define NHQ_ID   1  //!< command tag to obtain the module id
#define NHQ_W    2  //!< command tag for reading/setting the echo wait time
#define NHQ_U    3  //!< command tag for reading the measured hv
#define NHQ_I    4  //!< command tag for reading the measured current
#define NHQ_M    5  //!< command tag for reading the voltage limit value
#define NHQ_N    6  //!< command tag for reading the current limit value
#define NHQ_D    7  //!< command tag for reading/setting the demand hv 
#define NHQ_V    8  //!< command tag for reading/setting the hv ramping speed
#define NHQ_L    9  //!< command tag for reading/setting the current trip level
#define NHQ_S   10  //!< command tag for reading the status message
#define NHQ_T   11  //!< command tag for reading the status code
#define NHQ_A   12  //!< command tag for reading/setting auto startup state

//--- define output command tags -------------------------------------
#define NHQ_OUTPUT_CMD 0 //!< general output command tag for user_data.cmd
#define NHQ_OUTPUT_D1  1 //!< output command tag for demand voltage channel 1
#define NHQ_OUTPUT_D2  2 //!< output command tag for demand voltage channel 2
#define NHQ_OUTPUT_V1  3 //!< output command tag for hv ramping speed channel 1
#define NHQ_OUTPUT_V2  4 //!< output command tag for hv ramping speed channel 2

//--- define various timers ------------------------------------------
#define NHQ_MAX_READ  5 //!< read loop maximum
#define NHQ_ECHO_WAIT 3 //!< echo wait time in (ms)
#define NHQ_WAIT      7 //!< wait time between command and data in units of 10(ms)

//--- define of array limits -----------------------------------------
#define NHQ_MAX_STR   16

extern bit FREEZE_MODE;
extern bit DEBUG_MODE;

char code node_name[] = "SCS-210-NHQ";

/* declare number of sub-addresses to framework */
unsigned char idata _n_sub_addr = 1;

bit output_flag, flush_flag;
unsigned char n_in;

static unsigned long xdata start_time;              //!< timer for the 'wait between command and data' routine [nhq_wait()]
static unsigned long xdata cycle_timer_frequently;  //!< timer for cyclic readout of the 'urgend data', e.g. measured voltage
static unsigned long xdata cycle_timer_scarcely;    //!< timer for cyclic readout of the 'less urgend data', e.g. status
char xdata str[NHQ_MAX_STR];                        //!< global string in xdata to keep data memory space small

unsigned char nhq_read(unsigned char cmd, unsigned char channel);
unsigned char nhq_multiple_read(unsigned char cmd, unsigned char channel, unsigned char length, unsigned char trials);
void nhq_wait(unsigned char wait) reentrant;

/*---- Define channels and configuration parameters returned to
       the CMD_GET_INFO command                                 ----*/

/* data buffer (mirrored in EEPROM) */

struct {
   char          cmd[NHQ_MAX_STR];
   char          cmd_demand_voltage[2][NHQ_MAX_STR];
   char          cmd_hv_ramp[2][NHQ_MAX_STR];
   char          id[32];
   unsigned char wait_time;
   float         measured_voltage[2];
   float         measured_current[2];
   float         voltage_limit[2];
   float         current_limit[2];
   float         demand_voltage[2];
   float         hv_ramp[2];
   float         current_trip_level[2];
   char          status_msg_ch1[8];
   char          status_msg_ch2[8];
   unsigned char status_code[2];
   unsigned char auto_start[2];
   unsigned char baud;
} xdata user_data;

MSCB_INFO_VAR code variables[] = {
   1,   UNIT_ASCII,  0, 0, MSCBF_DATALESS, "RS232", 0,
   16,  UNIT_STRING, 0, 0,              0, "Cmd", &user_data.cmd[0],
   16,  UNIT_STRING, 0, 0,              0, "Cmd_D1", &user_data.cmd_demand_voltage[0][0],
   16,  UNIT_STRING, 0, 0,              0, "Cmd_D2", &user_data.cmd_demand_voltage[1][0],
   16,  UNIT_STRING, 0, 0,              0, "Cmd_V1", &user_data.cmd_hv_ramp[0][0],
   16,  UNIT_STRING, 0, 0,              0, "Cmd_V2", &user_data.cmd_hv_ramp[1][0],
   32,  UNIT_STRING, 0, 0,              0, "ID", &user_data.id[0],
   1,   UNIT_BYTE,   0, 0,              0, "WaitTime", &user_data.wait_time,
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "U1", &user_data.measured_voltage[0],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "U2", &user_data.measured_voltage[1],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "I1", &user_data.measured_current[0],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "I2", &user_data.measured_current[1],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "M1", &user_data.voltage_limit[0],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "M2", &user_data.voltage_limit[1],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "N1", &user_data.current_limit[0],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "N2", &user_data.current_limit[1],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "D1", &user_data.demand_voltage[0],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "D2", &user_data.demand_voltage[1],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "V1", &user_data.hv_ramp[0],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "V2", &user_data.hv_ramp[1],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "L1", &user_data.current_trip_level[0],
   4,   UNIT_WORD,   0, 0,    MSCBF_FLOAT, "L2", &user_data.current_trip_level[1],
   8,   UNIT_STRING, 0, 0,              0, "S1", &user_data.status_msg_ch1[0],
   8,   UNIT_STRING, 0, 0,              0, "S2", &user_data.status_msg_ch2[0],
   1,   UNIT_BYTE,   0, 0,              0, "T1", &user_data.status_code[0],
   1,   UNIT_BYTE,   0, 0,              0, "T2", &user_data.status_code[1],
   1,   UNIT_BYTE,   0, 0,              0, "A1", &user_data.auto_start[0],
   1,   UNIT_BYTE,   0, 0,              0, "A2", &user_data.auto_start[1],
   1,   UNIT_BAUD,   0, 0,              0, "Baud", &user_data.baud,
   0
};

/********************************************************************\

  Application specific init and input/output routines

\********************************************************************/

void user_write(unsigned char channel) reentrant;
//void write_gain(void);

/*---- User init function ------------------------------------------*/
/*!
 * <p> Initializes the user software part. It sets the baud rate,
 * than is reading back all the variables from the iseg NHQ module
 * to which it is connected. At the end it starts the cyclic timers, 
 * i.e. makes a first time stamp.
 *
 * \param init tag, when set, sets the baud rate to 9600 (default)
 */
void user_init(unsigned char init)
{
   /* initialize UART1 */
   if (init)
      user_data.baud = BD_9600;    // 9600 by default

   uart_init(1, user_data.baud);

   strcpy(user_data.cmd, "");
   strcpy(user_data.cmd_demand_voltage[0], "");
   strcpy(user_data.cmd_demand_voltage[1], "");
   strcpy(user_data.cmd_hv_ramp[0], "");
   strcpy(user_data.cmd_hv_ramp[1], "");

   watchdog_enable(2);

   nhq_multiple_read(NHQ_ID, 0, 23, NHQ_MAX_READ); // read nqh id
   nhq_multiple_read(NHQ_W,  0,  6, NHQ_MAX_READ); // read wait time
   nhq_multiple_read(NHQ_M,  1,  6, NHQ_MAX_READ); // read voltage limit of channel 1
   nhq_multiple_read(NHQ_M,  2,  6, NHQ_MAX_READ); // read voltage limit of channel 2
   nhq_multiple_read(NHQ_N,  1,  6, NHQ_MAX_READ); // read current limit of channel 1
   nhq_multiple_read(NHQ_N,  2,  6, NHQ_MAX_READ); // read current limit of channel 2
   nhq_multiple_read(NHQ_V,  1,  6, NHQ_MAX_READ); // read hv ramp speed of channel 1
   nhq_multiple_read(NHQ_V,  2,  6, NHQ_MAX_READ); // read hv ramp speed of channel 2
   nhq_multiple_read(NHQ_L,  1,  6, NHQ_MAX_READ); // read current trip level of channel 1
   nhq_multiple_read(NHQ_L,  2,  6, NHQ_MAX_READ); // read current trip level of channel 2
   nhq_multiple_read(NHQ_A,  1,  5, NHQ_MAX_READ); // read auto start status of channel 1
   nhq_multiple_read(NHQ_A,  2,  5, NHQ_MAX_READ); // read auto start status of channel 2

   nhq_multiple_read(NHQ_T,  1,  7, NHQ_MAX_READ); // read status of channel 1
   nhq_multiple_read(NHQ_T,  2,  7, NHQ_MAX_READ); // read status of channel 2
   nhq_multiple_read(NHQ_S,  1,  9, NHQ_MAX_READ); // read status message of channel 1
   nhq_multiple_read(NHQ_S,  2,  9, NHQ_MAX_READ); // read status message of channel 2

   nhq_multiple_read(NHQ_D,  1,  7, NHQ_MAX_READ); // read demand hv of channel 1
   nhq_multiple_read(NHQ_D,  2,  7, NHQ_MAX_READ); // read demand hv of channel 2
   nhq_multiple_read(NHQ_U,  1,  6, NHQ_MAX_READ); // read measured hv of channel 1
   nhq_multiple_read(NHQ_U,  2,  6, NHQ_MAX_READ); // read measured hv of channel 2
   nhq_multiple_read(NHQ_I,  1,  8, NHQ_MAX_READ); // read measured current of channel 1
   nhq_multiple_read(NHQ_I,  2,  8, NHQ_MAX_READ); // read measured current of channel 2

   // start readback timer
   cycle_timer_frequently = time();
   cycle_timer_scarcely   = time();
}

/*---- User write function -----------------------------------------*/

/* buffers in mscbmain.c */
extern unsigned char xdata in_buf[64], out_buf[64];

#pragma NOAREGS

char idata obuf[8];

/*--------------------------------------------------------------------*/
/*!
 * <p> Interrupt routine in case the user is sending anything from the
 * computer via the mscb bus. Since it is an interrupt routine the code
 * is very slim, and the main functionality is transfered via global
 * tags to the user_loop routine.
 *
 * \param index scs_210 variable index, see the variable 'variables'
 */
void user_write(unsigned char index) reentrant
{
   // set output flag for the user_loop routine in order to write a command
   switch (index) {
     case 28: // baud rate
       uart_init(1, user_data.baud);
	   break;
     default:
	   break;
   }
}

/*---- User read function ------------------------------------------*/
/*!
 * <p> User read routine. No functionality at the moment.
 *
 * \param channel scs_210 variable index, see the variable 'variables',  
 *                For the moment only a dummy.
 */
unsigned char user_read(unsigned char channel)
{
   if (channel);  // keeping the compiler quiet
   return 0;
}

/*--------------------------------------------------------------------*/
/*!
 * <p> User function called via CMD_USER command.
 *
 * \param data_in
 * \param data_out
 */
unsigned char user_func(unsigned char *data_in, unsigned char *data_out)
{
   /* echo input data */
   data_out[0] = data_in[0];
   data_out[1] = data_in[1];
   return 2;
}

/*--------------------------------------------------------------------*/
/*!
 * <p> Routine with an approximative timer which can be interrupted.
 * The time is given in cycles. 1 cycle approx 10 ms
 *
 * \param wait time in units of 10 ms.
 */
void nhq_wait(unsigned char wait) reentrant
{
  start_time = time();
  do {
  } while (time()-start_time < wait);
}

/*--------------------------------------------------------------------*/
/*!
 * <p>Handles the actual communication between SCS210 and the NHQ.
 * It works as following: (i) a command string is sent, byte by byte.
 * After sending a byte SCS210->NHQ, the NHQ sends back an echo after
 * a delay of 3 ms. After this ping-pong, the interesting information
 * is read back, again byte by byte with a wating time between them.
 */
void nhq_communication()
{
  unsigned char len;
  unsigned char i;
  char          c;

  // send read command and eat up the echo
  len = strlen(str); // str is a global variable!!
  for (i=0 ; i<len ; i++) {
    putchar(str[i]);         // write a byte
    delay_ms(NHQ_ECHO_WAIT); // wait
    c = getchar_nowait();    // read back echo byte
    flush();
  }

  // wait a bit before reading back
  nhq_wait(NHQ_WAIT);
  flush(); 

  // read back info
  strcpy(str, "!!!!");
  i = 0;
  do {
    c = getchar_nowait();
    delay_ms(NHQ_ECHO_WAIT); // wait
    if (c != -1 && i < NHQ_MAX_STR) 
      str[i++] = c;
	if (c == '\n')  // end of data found
	  c = -1;
  } while (c != -1 && i < NHQ_MAX_STR);
  str[i] = '\0'; // terminate string
  flush();

  // feed watchdog
  watchdog_refresh(0);
}

/*--------------------------------------------------------------------*/
/*!
 * <p> Reads all the various variables from the iseg NHQ20xm module. 
 *
 * \param cmd     command tag
 * \param channel hv channel (1 or 2)
 */
unsigned char nhq_read(unsigned char cmd, unsigned char channel) 
{
  int xdata     i, value;
  float xdata   fvalue;

  // generate command string
  switch (cmd) {
    case NHQ_SYNC:
	  strcpy(str, "\r\n");
	  break;
    case NHQ_ID:
      strcpy(str, "#\r\n");
 	  break; 
    case NHQ_W:
      strcpy(str, "W\r\n");
	  break; 
    case NHQ_U:
      sprintf(str, "U%d\r\n", (int)channel);
	  break; 
    case NHQ_I:
      sprintf(str, "I%d\r\n", (int)channel);
	  break; 
    case NHQ_M:
      sprintf(str, "M%d\r\n", (int)channel);
   	  break; 
    case NHQ_N:
      sprintf(str, "N%d\r\n", (int)channel);
	  break; 
    case NHQ_D:
      sprintf(str, "D%d\r\n", (int)channel);
	  break; 
    case NHQ_V:
      sprintf(str, "V%d\r\n", (int)channel);
      break; 
    case NHQ_L:
      sprintf(str, "L%d\r\n", (int)channel);
	  break; 
    case NHQ_S:
      sprintf(str, "S%d\r\n", (int)channel);
   	  break; 
    case NHQ_T:
      sprintf(str, "T%d\r\n", (int)channel);
	  break; 
    case NHQ_A:
      sprintf(str, "A%d\r\n", (int)channel);
	  break; 
 	default:
	  strcpy(str, "\r\n");
	  break;
  }

  nhq_communication();

  // analyze read back
  switch (cmd) {
    case NHQ_ID:
	  if (i>2) 
  	    str[i-2] = '\0'; // get rid of '\r\n'
      strcpy(user_data.id, str);
 	  break; 
    case NHQ_W:
      sscanf(str, "%d\r\n", &value);
      user_data.wait_time = (unsigned char)value;
	  break; 
    case NHQ_U:
	  if ((strlen(str)!=7) && ((str[0]!='+') || (str[0]!='-'))) { // check if valid reading str should have the form '+nnnn\r\n' or '-nnnn\r\n'
	    i=-1;
	    break;
	  }
      sscanf(str, "%f\r\n", &fvalue);
	  user_data.measured_voltage[channel-1] = fvalue;
   	  break; 
    case NHQ_I:
	  if (str[4]!='-') // check if valid reading: str should have the form 'nnnn-n\r\n'
	    break;
      sscanf(str, "%d-%d\r\n", &value, &i);       // get mantissa and exponent
	  sprintf(str, "%0.1fE-%d", (float)value, i); // construct a 'float string'
	  sscanf(str, "%f", &fvalue);				  // extract float from 'float string'
	  user_data.measured_current[channel-1]=fvalue;
	  break; 
    case NHQ_M:
      sscanf(str, "%f\r\n", &fvalue);
	  user_data.voltage_limit[channel-1]=fvalue;
   	  break; 
    case NHQ_N:
      sscanf(str, "%f\r\n", &fvalue);
	  user_data.current_limit[channel-1]=fvalue;
	  break; 
    case NHQ_D:
      sscanf(str, "%f\r\n", &fvalue);
	  user_data.demand_voltage[channel-1]=fvalue;
	  break; 
    case NHQ_V:
      sscanf(str, "%f\r\n", &fvalue);
	  user_data.hv_ramp[channel-1]=fvalue;
	  break; 
    case NHQ_L:
      sscanf(str, "%f\r\n", &fvalue);
	  user_data.current_trip_level[channel-1]=fvalue;
	  break; 
    case NHQ_S:
	  if (i>2)
	    str[i-2]='\0';
      if (channel == 1)
        strncpy(user_data.status_msg_ch1, str, 8);
	  else
        strncpy(user_data.status_msg_ch2, str, 8);
	  break; 
    case NHQ_T:
      sscanf(str, "%d\r\n", &value);
      user_data.status_code[channel-1] = (unsigned char)value;
	  break; 
    case NHQ_A:
      sscanf(str, "%d\r\n", &value);
      user_data.auto_start[channel-1] = (unsigned char)value;
	  break; 
    default:
	  break;
  }
  return i;
}

/*--------------------------------------------------------------------*/
/*!
 * <p> loop read routine. Since the iseg software interface is really stupid,
 * the read back is looped a few times in order to increase read back failiures.
 *
 * \param cmd	  command tag
 * \param channel hv channel (1 or 2)
 * \param length  expected length of the read back string
 * \param trials  no of trials
 */
unsigned char nhq_multiple_read(unsigned char cmd, unsigned char channel, unsigned char length, unsigned char trials)
{
   unsigned char i, j=0;

   do {
     i=nhq_read(cmd, channel);
     j++;
     nhq_wait(NHQ_WAIT-2);
   } while ((i!=length) && (j<trials));
   flush();

   return j;
}

/*--------------------------------------------------------------------*/
/*!
 * <p> iseg write command handler.
 */
void nhq_handle_cmd(unsigned char output_cmd)
{
  // append command termination to command string
  switch (output_cmd) {
    case NHQ_OUTPUT_CMD:
	  sprintf(str, "%s%s", user_data.cmd, "\r\n");
	  break;
	case NHQ_OUTPUT_D1:
      sprintf(str, "%s%s", user_data.cmd_demand_voltage[0], "\r\n");
	  break;
	case NHQ_OUTPUT_D2:
      sprintf(str, "%s%s", user_data.cmd_demand_voltage[1], "\r\n");
	  break;
	case NHQ_OUTPUT_V1:
      sprintf(str, "%s%s", user_data.cmd_hv_ramp[0], "\r\n");
	  break;
	case NHQ_OUTPUT_V2:
      sprintf(str, "%s%s", user_data.cmd_hv_ramp[1], "\r\n");
	  break;
	default:
	  return;
  }

  nhq_communication();

  // --- D1, D2 related checks ---------------------------------------------------------------------------

  // if cmd is D1/D2 activate it
  if (NHQ_OUTPUT_CMD && (strstr(user_data.cmd, "D1=") || strstr(user_data.cmd, "D2="))) { // cmd =D1, D2
    if (strstr(user_data.cmd, "D1="))
	  strcpy(str, "G1\r\n");
	else
	  strcpy(str, "G2\r\n");

	nhq_communication();
  }
  
  // Cmd_D1 or Cmd_D2
  if (NHQ_OUTPUT_D1 || NHQ_OUTPUT_D2) { 
    if (NHQ_OUTPUT_D1)
	  strcpy(str, "G1\r\n");
	else
	  strcpy(str, "G2\r\n");

	nhq_communication();
  }
}

/*--------------------------------------------------------------------*/
/*!
 * <p> Filter to ensure that only acceptable commands are reaching the
 * iseg module.
 */
void nhq_write(unsigned char output_cmd)
{
  switch (output_cmd) {
    case NHQ_OUTPUT_CMD:
	  if (strstr(user_data.cmd, "W=") || strstr(user_data.cmd, "D1=") || strstr(user_data.cmd, "D2=") ||
	      strstr(user_data.cmd, "G1=") || strstr(user_data.cmd, "G2=") || strstr(user_data.cmd, "V1=") || strstr(user_data.cmd, "V2=") ||
		  strstr(user_data.cmd, "L1=") || strstr(user_data.cmd, "L2=") || strstr(user_data.cmd, "A1=") || strstr(user_data.cmd, "A2="))
	    nhq_handle_cmd(output_cmd);
	  break;
    case NHQ_OUTPUT_D1:
	  if (strstr(user_data.cmd_demand_voltage[0], "D1=")) // check if expected command
	    nhq_handle_cmd(output_cmd);
	  break;
    case NHQ_OUTPUT_D2:
	  if (strstr(user_data.cmd_demand_voltage[1], "D2=")) // check if expected command
	    nhq_handle_cmd(output_cmd);
	  break;
    case NHQ_OUTPUT_V1:
	  if (strstr(user_data.cmd_hv_ramp[0], "V1=")) // check if expected command
	    nhq_handle_cmd(output_cmd);
	  break;
    case NHQ_OUTPUT_V2:
	  if (strstr(user_data.cmd_hv_ramp[1], "V2=")) // check if expected command
	    nhq_handle_cmd(output_cmd);
	  break;
    default:
	  break;
  }
}

/*---- User loop function ------------------------------------------*/

/* external watchdog */
sbit EWD = P0 ^ 5;

/*!
 * <p> Main user loop. Reads cyclicly the relevant variables and handles user
 * commands.
 */
void user_loop(void)
{
   if (flush_flag) {
      flush_flag = 0;
      flush();
   }

   if (time() - cycle_timer_frequently > 100) { // read back every 1 sec
     nhq_multiple_read(NHQ_U,  1,  6, NHQ_MAX_READ); // read measured hv of channel 1
     nhq_multiple_read(NHQ_I,  1,  8, NHQ_MAX_READ); // read measured current of channel 1
     nhq_multiple_read(NHQ_D,  1,  7, NHQ_MAX_READ); // read demand hv of channel 1
     nhq_multiple_read(NHQ_U,  2,  6, NHQ_MAX_READ); // read measured hv of channel 2
     nhq_multiple_read(NHQ_I,  2,  8, NHQ_MAX_READ); // read measured current of channel 2
     nhq_multiple_read(NHQ_D,  2,  7, NHQ_MAX_READ); // read demand hv of channel 2
   }

   if (time() - cycle_timer_scarcely > 500) { // read back every 5 sec
     nhq_multiple_read(NHQ_T,  1,  7, NHQ_MAX_READ); // read status of channel 1
     nhq_multiple_read(NHQ_S,  1,  9, NHQ_MAX_READ); // read status message of channel 1
     nhq_multiple_read(NHQ_V,  1,  6, NHQ_MAX_READ); // read hv ramp speed of channel 1
     nhq_multiple_read(NHQ_L,  1,  6, NHQ_MAX_READ); // read current trip level of channel 1
     nhq_multiple_read(NHQ_A,  1,  5, NHQ_MAX_READ); // read auto start status of channel 1
     nhq_multiple_read(NHQ_T,  2,  7, NHQ_MAX_READ); // read status of channel 2
     nhq_multiple_read(NHQ_S,  2,  9, NHQ_MAX_READ); // read status message of channel 2
     nhq_multiple_read(NHQ_V,  2,  6, NHQ_MAX_READ); // read hv ramp speed of channel 2
     nhq_multiple_read(NHQ_L,  2,  6, NHQ_MAX_READ); // read current trip level of channel 2
     nhq_multiple_read(NHQ_A,  2,  5, NHQ_MAX_READ); // read auto start status of channel 2
   }
   
   if (strlen(user_data.cmd)>0) { // cmd via msc and midas DD
     nhq_write(NHQ_OUTPUT_CMD);
     strcpy(user_data.cmd, ""); // reset cmd
   }

   if (strlen(user_data.cmd_demand_voltage[0])>0) { // set demand voltage of ch1, via msc and midas DD
     nhq_write(NHQ_OUTPUT_D1);
     strcpy(user_data.cmd_demand_voltage[0], ""); // reset cmd
   }

   if (strlen(user_data.cmd_demand_voltage[1])>0) { // set demand voltage of ch2, via msc and midas DD
     nhq_write(NHQ_OUTPUT_D2);
     strcpy(user_data.cmd_demand_voltage[1], ""); // reset cmd
   }

   if (strlen(user_data.cmd_hv_ramp[0])>0) { // set hv ramping of ch1, via msc and midas DD
     nhq_write(NHQ_OUTPUT_V1);
     strcpy(user_data.cmd_hv_ramp[0], ""); // reset cmd
   }

   if (strlen(user_data.cmd_hv_ramp[1])>0) { // set hv ramping of ch2, via msc and midas DD
     nhq_write(NHQ_OUTPUT_V2);
     strcpy(user_data.cmd_hv_ramp[1], ""); // reset cmd
   }

/*
   // toggle external watchdog 
   EWD = !EWD;
*/

   // go into idle mode, wakeup via UART of 100Hz interrupt
   PCON |= 0x01;
}
